@@ -1,5 +1,6 @@
 from rwkvstic.agnostic.backends.base import module
 from typing import Dict
+import tensorflow as tf
 
 
 def LegacyRWKV(ops: module, *args):
@@ -8,8 +9,10 @@ def LegacyRWKV(ops: module, *args):
         @ ops.initfunc
         def __init__(self, w: Dict[str, ops.TensorType]):
             super(myRWKV, self).__init__()
+            print("Legacy RWKV")
 
-            self.ops = ops
+            for x in ops.__dict__.keys():
+                self.__dict__[x] = ops.__dict__[x]
             self.postprocess0: ops.VectorType = (w["ln_out.weight"])
             self.postprocess1: ops.VectorType = (w["ln_out.bias"])
             self.postprocess2: ops.VectorType = (w["head.weight"])
@@ -54,61 +57,90 @@ def LegacyRWKV(ops: module, *args):
                 w[f"blocks.{x}.ffn.value.weight"] for x in range(ops.n_layers)])
 
         @ops.layerdef
-        def doLayer(self, x, statea, stateb, statec, stated, statee, xx):
+        def doLayer(self, x, statea, stateb, statec, stated, statee, xx: int):
 
-            xy = ops.layernorm(x, self.ln1w[xx], self.ln1b[xx])
+            xy = self.layernorm(x, self.ln1w[xx], self.ln1b[xx])
 
-            k = ops.matvec(
-                self.key[xx], ops.lerp(statea, xy, self.kktk[xx]))
+            tc = self.push(self.roll(xy), statea)
 
-            v = ops.matvec(self.value[xx], ops.lerp(
-                statea, xy, self.vvtv[xx]))
+            k = self.matvec(
+                self.key[xx], self.lerp(tc, xy, self.kktk[xx]))
 
-            r = ops.logistical(ops.matvec(
-                self.receptance[xx], ops.lerp(statea, xy, self.rrtr[xx])))
+            v = self.matvec(self.value[xx], self.lerp(
+                tc, xy, self.vvtv[xx]))
 
-            ww = ops.add(k, self.time_first[xx])
-            p = ops.maximum(statee, ww)
+            rr = self.matvec(
+                self.receptance[xx], self.lerp(tc, xy, self.rrtr[xx]))
 
-            e1 = ops.exp(ops.subtract(statee, p))
-            e2 = ops.exp(ops.subtract(ww, p))
-            a = ops.add(ops.multiply(e1, stateb), ops.multiply(e2, v))
-            b = ops.add(ops.multiply(e1, statec), e2)
-            ww = ops.add(statee, self.time_decay[xx])
+            r = self.logistical(rr)
 
-            p = ops.maximum(ww, k)
+            ww = self.add(k, self.time_first[xx])
+            rz = self.emptyarray(self.len(x))
+            bz = self.emptyarray(self.len(x)+1)
+            cz = self.emptyarray(self.len(x)+1)
+            gz = self.emptyarray(self.len(x)+1)
+            bz = self.arrayPush(bz, statee, 0)
+            cz = self.arrayPush(cz, stateb, 0)
+            gz = self.arrayPush(gz, statec, 0)
 
-            e1 = ops.exp(ops.subtract(ww, p))
-            e2 = ops.exp(ops.subtract(k, p))
-            outb = ops.add(ops.multiply(e1, stateb), ops.multiply(e2, v))
-            outc = ops.add(ops.multiply(e1, statec), e2)
-            eee = p
-            wkv = ops.divide(a, b)
+            for i in self.rng(self.len(x)):
 
-            mvv = ops.add(x, ops.matvec(
-                self.outputvv[xx], ops.multiply(r, wkv)))
+                p = self.maximum(self.arrayGet(bz, i), ww[i])
 
-            ddd = ops.layernorm(mvv, self.ln2w[xx], self.ln2b[xx])
+                e1 = self.exp(self.subtract(self.arrayGet(bz, i), p))
 
-            km = ops.relu(ops.matvec(self.key_ffn[xx], ops.lerp(
-                stated, ddd, self.time_mix_k_ffn[xx])))
+                e2 = self.exp(self.subtract(ww[i], p))
 
-            rt = ops.logistical(ops.matvec(self.receptance_ffn[xx], ops.lerp(
-                stated, ddd, self.time_mix_r_ffn[xx])))
+                a = self.add(self.multiply(e1, self.arrayGet(cz, i)),
+                             self.multiply(e2, v[i]))
 
-            x = ops.add(mvv, ops.multiply(
-                ops.matvec(self.value_ffn[xx], ops.multiply(km, km)), rt))
+                b = self.add(self.multiply(e1, self.arrayGet(gz, i)), e2)
 
-            return x, xy, outb, outc, ddd, eee
+                wwn = self.add(self.arrayGet(bz, i), self.time_decay[xx])
+
+                p1 = self.maximum(wwn, k[i])
+
+                e11 = self.exp(self.subtract(wwn, p1))
+
+                e21 = self.exp(self.subtract(k[i], p1))
+
+                outb = self.add(self.multiply(e11, self.arrayGet(cz, i)),
+                                self.multiply(e21, v[i]))
+
+                outc = self.add(self.multiply(e11, self.arrayGet(gz, i)), e21)
+
+                wkv = self.divide(a, b)
+                rz = self.arrayPush(rz, wkv, i)
+                bz = self.arrayPush(bz, p1, i+1)
+                cz = self.arrayPush(cz, outb, i+1)
+
+                gz = self.arrayPush(gz, outc, i+1)
+
+            mvv = self.add(x, self.matvec(
+                self.outputvv[xx], self.multiply(r, self.stack(rz))))
+
+            ddd = self.layernorm(mvv, self.ln2w[xx], self.ln2b[xx])
+
+            rc = self.push(self.roll(ddd), stated)
+
+            km = self.relu(self.matvec(self.key_ffn[xx], self.lerp(
+                rc, ddd, self.time_mix_k_ffn[xx])))
+
+            rt = self.logistical((self.matvec(self.receptance_ffn[xx], self.lerp(
+                rc, ddd, self.time_mix_r_ffn[xx]))))
+
+            rvm = self.matvec(self.value_ffn[xx], self.multiply(km, km))
+
+            x = self.add(mvv, self.multiply(
+                rvm, rt))
+
+            return x,  self.pop(xy), self.arrayGet(cz, self.len(x)), self.arrayGet(gz, self.len(x)),  self.pop(ddd), self.arrayGet(bz, self.len(x))
 
         @ ops.mainfunc
-        def forward(self, x: ops.VectorType, state: ops.MatrixType = None):
-
-            if (state is None):
-                state = ops.emptyState
-
-            x = ops.layernorm(
-                ops.processEmbed(ops.getIndex(self.emb, x)),
+        def forward(self, x, state):
+            g = self.getIndex(self.emb, x)
+            x = self.layernorm(
+                self.processEmbed(g),
                 self.emb1, self.emb2)
 
             statea = state[0::5]
@@ -117,17 +149,23 @@ def LegacyRWKV(ops: module, *args):
             stated = state[3::5]
             statee = state[4::5]
 
-            ot = []
+            ot = self.emptyarray(self.n_layers*5)
+
+            for i in self.rng(self.n_layers):
 
-            for i in range(ops.n_layers):
                 x, aaa, bbb, ccc, ddd, eee = self.doLayer(
                     x, statea[i], stateb[i], statec[i], stated[i], statee[i], i)
-                ot = ot + [aaa, bbb, ccc, ddd, eee]
 
-            x = ops.matvec(self.postprocess2, ops.layernorm(x, self.postprocess0,
-                                                            self.postprocess1))
+                ot = self.arrayPush(ot, aaa, i*5)
+                ot = self.arrayPush(ot, bbb, i*5+1)
+                ot = self.arrayPush(ot, ccc, i*5+2)
+                ot = self.arrayPush(ot, ddd, i*5+3)
+                ot = self.arrayPush(ot, eee, i*5+4)
 
-            return ops.postProcessTensor(x), ops.stack(ot)
+            x = self.matvec(self.postprocess2, self.layernorm(x, self.postprocess0,
+                                                              self.postprocess1))
+
+            return self.postProcessTensor(self.pop(x)), self.stack(ot)
 
         # for keras stuff, ignore this if you are not using keras
         def call(self, *args, **kwds):
